/*
 * 2005 T W Bennet
 *
 * Use the Christophe Devine DES code to encrypt or decript a file in
 * CBC mode (presuming I've implemented the mode correctly).  The DES
 * itself I have crippled to use a 16-bit key.
 *
 * cripcryp [ -d | -e ] key [ fn ]
 *
 * Uses the key to encrypt (-e, default) or decrypt (-d) the file fn, or
 * standard input.  Writes the result to standard output.  If the key starts
 * with 0x, it is taken as a sixteen-bit hex number.  Otherwise, the ascii
 * values of the characters are used.  The first three are appended.  If a
 * fourth is provided, its low bits are used to fill in the high bits of the
 * other three characters.
 *
 * The encryption algorithm computes a simple 16-bit checksum which is added
 * to the data before encryption, or checked by the decryption alg.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include "cripdes.h"

/* The Initialization Vector (IV). */
#define IV { 0x45, 0x12, 0x9f, 0xa4, 0x31, 0xcd, 0x98, 0x31 }

static void usage(char *pgm)
{
	fprintf(stderr, "Usage: %s [ -d | -e ] [ -q ] key [ infile ]\n", pgm);
	exit(1);
}

/* Add to the checksum. */
static void addchk(uint16 *sum, uint8 *data, int size)
{
	uint32 term = *sum;

	while(size) {
		term += (*data++ & 0x0ff) << 8;
		if(--size) {
			term += (*data++ & 0x0ff);
			--size;
		}
	}

	while(term & 0xffff0000)
		term = (term & 0x0ffff) + (term >> 16);

	*sum = term;
}

/* XOR the first block in place with the second. */
static void xor(uint8 targ[8], uint8 vec[8])
{
#ifdef BUFTEST
	memcpy(targ, vec, sizeof(uint8 [8]));
#else
	int m = 8;
	while(m--) *targ++ ^= *vec++;
#endif
}

static FILE *in;	/* Input stream. */

/* Read a block and croak on error. */
static int getblk(uint8 blk[8])
{
	int m = fread(blk, sizeof(uint8), 8, in);
	if(ferror(in)) {
		fprintf(stderr, "Read failed: %s.\n", strerror(errno));
		exit(1);
	}

	return m;
}

/* For decode reads, all blocks must be complete. */
static int sizeok(int sz)
{
	if(sz != 8 && sz != 0) {
		fprintf(stderr, "Short block (%d) on decode read.\n", sz);
		exit(1);
	}

	return sz;
}

/* Encryption reader.  It will keep track of the checksum and add it
   to the end of the stream.  It will add three bytes, perhaps with some
   zero padding.  The last three bytes before eof are the checksum and
   the total number bytes (including the checksum) which extend beyond
   the actual file file size.  If the three bytes wont' fit in the last
   block, an additional block is added.  The additional bytes are always
   at the end of whatever block they occupy. */
static int getclear(uint8 blk[8])
{
	static uint16 csum = 0;
	static char done = 0;
	static int morefluff = 0;
	int m;

	if(done) return 0;

	/* Do a plain read.  If there's enough data, return. */
	m = getblk(blk);
	addchk(&csum, blk, m * sizeof(uint8));
	if(m == 8) return 1;

	/* Okay.  See if we need to add stuff. */
	if(m > 5) {
		morefluff = 8 - m;
		memset(blk + m, 0, morefluff * sizeof(uint8));
		return 1;
	} 

	/* Add the extra information .*/
	blk[5] = (8 - m + morefluff) * sizeof(uint8);
	blk[6] = (csum >> 8) & 0x0ff;
	blk[7] = csum & 0x0ff;
	memset(blk + m, 0, (8 - m - 3) * sizeof(uint8));
	done = 1;
	return 1;
}

/* Write the stdout with error msg. */
void genblk(uint8 blk[8], int count)
{
	if(fwrite(blk, sizeof(uint8), count, stdout) < count) {
		fprintf(stderr, "Write failed: %s.\n", strerror(errno));
		exit(1);
	}
}

/* Check for zero. */
int iszero(uint8 *start, int cnt)
{
	while(cnt--) if(*start++) return 0;
	return 1;
}

/* Output buffering for the decode stream.  Writes two behind, since the
   file size may not be known for a while.  Call with NULL to flush and 
   verify. */
void putclear(uint8 blk[8])
{
	static uint8 buf[2][8];
	static int next = 0;	/* Next to go out. */
	static int used = 0;	/* Number filled. */
	static uint16 csum = 0; /* Checksum of file data. */

	/* Flush case. */
	if(blk == NULL) {
		int last = (next + used - 1) % 2; /* Location of last block*/
		uint16 scsum = 0;	/* Stored sum. */
		int lose;		/* Bytes to lose. */

		/* Empty encry files are illegal. */
		if(used == 0) {
			fprintf(stderr, "Illegal empty encrypted file.\n");
			exit(1);
		}

		/* Extract the control info. */
		lose = buf[last][5];
		scsum = ((buf[last][6] << 8) & 0x0ff00)|(buf[last][7] & 0x0ff);

		/* Check for too much loss. */
		if(lose > 10 || (lose > 8 && used < 2)) {
			fprintf(stderr, "Unreasonable stored file size.  "
				"Most likely a bad decrypt key.\n");
			exit(1);
		}
			
		/* Check for zero bytes in the filler. */
		if(lose > 3) {
			int ok;
			if(lose >= 8) {
				ok = iszero(buf[1 - next], 5);
				if(ok && lose > 8 && 
				   !iszero(buf[next] + (16 - lose),
					   lose - 8))
					ok = 0;
			} else
				ok = iszero(buf[(next + (used - 1)) % 2] +
					    8 - lose, lose - 3);
			if(!ok) {
				fprintf(stderr, 
					"Non-zero fill bytes.  "
					"Most likely a bad decrypt key.\n");
				exit(1);
			}
		}

		/* If last block is loss, lose it. */
		if(lose >= 8) {
			lose -= 8;
			used--;
		}

		/* If still two, gen first one. */
		if(used > 1) {
			/* First pending buffer is all data. */
			genblk(buf[next], 8);
			addchk(&csum, buf[next], 8 * sizeof(uint8));
			next = 1 - next;
			used--;
		}

		/* If any are left, gen them. */
		if(used) {
			genblk(buf[next], 8 - lose);
			addchk(&csum, buf[next], (8 - lose) * sizeof(uint8));
		}

		/* Check the checksum. */
		if(csum != scsum) {
			fprintf(stderr, "Checksum error on decrypted file.\n");
			exit(1);
		}

		return;
	}

	/* Normal case. */
	if(used == 2) {
		genblk(buf[next], 8);
		addchk(&csum, buf[next], 8 * sizeof(uint8));
		memcpy(buf[next], blk, sizeof buf[next]);
		next = 1 - next;
		return;
	} else 
		/* Filling up case. */
		memcpy(buf[used++], blk, sizeof buf[0]);
}

#ifdef BUFTEST
int  des_set_key( des_context *ctx, uint8 key[8] ) { }
void des_encrypt( des_context *ctx, uint8 input[8], uint8 output[8] )
{
	memcpy(output, input, sizeof(uint8[8]));
}
void des_decrypt( des_context *ctx, uint8 input[8], uint8 output[8] )
{
	memcpy(output, input, sizeof(uint8[8]));
}
#endif

int main(int argc, char **argv)
{
	char *key;		/* User key. */
	des_context ctx;	/* Key context for DES. */
	int m;			/* General purpose. */
	uint16 chksum;		/* Checksum. */

	/* Options, all three of them. */
	char enc = 0, dec = 0;
	while(1) {
		int op = getopt(argc, argv, "deq");
		if(op < 0) break;
		switch(op) {
		case 'd':
			dec = 1;
			break;
		case 'e':
			enc = 1;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	/* Flag rules and default. */
	if(enc && dec) {
		fprintf(stderr, "%s: Sorry, can't both encode and decode.\n",
			argv[0]);
		exit(1);
	}
	if(!enc && !dec) enc = 1;

	/* Key. */
	if(optind >= argc) usage(argv[0]);
	key = argv[optind++];
	m = strlen(key);
	if(m < 3) {
		fprintf(stderr, "Keys must be at least 3 characters.\n");
		exit(4);
	} 

	/* Opt filename. */
	if(optind < argc) {
		if(argc > optind + 1) usage(argv[0]);
		if(strcmp(argv[optind], "-") == 0)
			in = stdin;
		else if((in = fopen(argv[optind], "rb")) == NULL) {
			fprintf(stderr, "%s: Cannot read %s: %s.\n",
				argv[0], argv[optind], strerror(errno));
			exit(1);
		}
	} else
		in = stdin;

	/* Compose the key. */
	if(key[0] == '0' && key[1] == 'x') {
		/* Hex number. */
		unsigned int hkey;
		sscanf(key + 2, "%x", &hkey);
		key[0] = (hkey >> 16) & 0x0ff;
		key[1] = (hkey >> 8) & 0x0ff;
		key[2] = hkey & 0x0ff;
	} else {
		/* A third plain digit can be used to add high bits. */
		key[0] |= (key[3] << 7) & 0x080;
		key[1] |= (key[3] << 6) & 0x080;
		key[2] |= (key[3] << 5) & 0x080;
	}
	des_set_key(&ctx, key);

	/* Do the translation. */
	if(enc) {
		uint8 inblk[8];
		uint8 outblk[8] = IV;

		/* Read the file, writing encrypted blocks. */
		while(getclear(inblk)) {
			/* Encrypt */
#ifndef BUFTEST
			xor(inblk, outblk);
#endif
			des_encrypt(&ctx, inblk, outblk);

			/* Write the block. */
			genblk(outblk, 8);
		}
	} else {
		uint8 buf1[8] = IV;
		uint8 buf2[8];

		uint8 *inblk = buf2;
		uint8 *outblk = buf1;

		while(1) {
			uint8 tmp[8];	/* Temp block. */
			uint8 *t;	/* Temp ptr for swap. */

			/* Read the segment. */
			m = sizeok(getblk(inblk));
			if(m < 8) break;

			/* Decrypt. */
			des_decrypt(&ctx, inblk, tmp);
			xor(outblk, tmp);

			/* Write the block. */
			putclear(outblk);

			/* Swap these. */
			t = inblk;
			inblk = outblk;
			outblk = t;
		}
		putclear(NULL);
	}
}
